#!/usr/bin/env python3

import json
import logging
import re

import eccs_properties as e_p
from flask import Flask, request, jsonify
from flask_restful import Resource, Api
from utils import get_logger, get_list_from_url, get_reg_auth_dict, generate_login_url
from markupsafe import escape
from datetime import timedelta, datetime, date
from urllib.parse import urlparse, parse_qs, unquote

app = Flask(__name__)
api = Api(app)


### Functions

# Build Email Addresses Link for ECCS Web Gui
def buildEmailAddress(listContacts):
    listCtcs = listContacts.split(",")
    hrefList = []

    for email in listCtcs:
       hrefList.append(f"<a href='{email}'>{email.replace('mailto:', '')}</a>")
 
    return hrefList

# Research the value of the research_item into ECCS output files
def existsInFile(file_path, value, research_item, eccsDataTable, date):
    try:
       with open(file_path,"r",encoding="utf-8") as fo:
            lines = fo.readlines()
    except FileNotFoundError as e:
        if (eccsDataTable):
           return ''
        else:
           return jsonify(error=f'FileNotFound: ECCS script has not been executed on {date} yet')

    for line in lines:
        aux = json.loads(line)
        if (research_item == "entityID"):
           if (value == aux['entityID']):
              return True
        if (research_item == "registrationAuthority"):
           if (value == aux['registrationAuthority']):
              return True
    return False

def getSimpleDict(aux):
    simpleDict = {
       "date": aux['date'],
       "displayName": aux['displayName'],
       "entityID": aux['entityID'],
       "registrationAuthority": aux['registrationAuthority'],
       "status": aux['status'],
       "checkResult": [aux["sp1"]["checkResult"],aux["sp2"]["checkResult"],aux["sp3"]["checkResult"]]
    }
    return simpleDict

def isValidDate(date_text):
      try:
         date.fromisoformat(date_text)
      except ValueError as e:
         print(e)
         return False
      return True

def clearDateString(str):
    del_char = {"[","]"}
    for elem in del_char:
        str = str.replace(elem, '')
    return str

def storeParsedDay(path, data):
   try: 
      with open(path, "w") as outfile:
         outfile.write(json.dumps(data))
   except FileNotFoundError as e:
      print(e)

# Log will be parsed using predefined format
# %(addr)|[%(ctime)]|%(method)|%(uri)|%(uagent)|%(referer)
# target result is array like: 
# [
# "YYYY-MM-DD" => [
#     "request_count" => [
#        web => 100,
#        api => 200 
#     ],
#     "requested_param" => [
#        idp => 100,
#        reg_auth => 200 
#     ],
#     idp => [
#        "xx" => [
#           "api" => 1,
#           "web" => 10
#        ],
#     ],
#     reg_auth => [
#        "xx" => [
#           "api" => 1,
#           "web" => 10
#        ],
#      ],
# ],
#]
def parseLog(lines,criteria):
    result = {}
    request_count = {'web':0,'api':0}
    request_param = {'idp':0,'reg_auth':0}
    idp = {}
    reg_auth = {}
    rowDateCheck = ''
   
    for line in lines:
        row = line.split('|')
        
        if len(row) <= 1 or row[3].find("api/webdata") >= 0:
            continue
        
        rowDate = datetime.strptime(clearDateString(row[1]), '%a %b %d %H:%M:%S %Y').strftime('%Y-%m-%d')
        
      # check for entries to be in range of date of rotated log  
        if 'cur_date' in criteria and criteria['cur_date'] != rowDate:
            continue
         
        if ('date_from' in criteria and 'date_to' in criteria and criteria['date_from'] <= rowDate <= criteria['date_to']) or ('date_from' not in criteria and 'date_to' not in criteria and rowDate):
            
            if rowDateCheck != rowDate:
                request_count = {'web':0,'api':0}
                request_param = {'idp':0,'reg_auth':0}
                idp = {}
                reg_auth = {}
                rowDateCheck = rowDate
                        
            rowGET = parseReqURL(row[3])
            
            if ('idp' not in criteria and 'reg_auth' not in criteria) or ('idp' in criteria and rowGET['idp'] and rowGET['idp'].find(criteria['idp']) >= 0) or ('reg_auth' in criteria and rowGET['reg_auth'] and rowGET['reg_auth'].find(criteria['reg_auth']) >= 0):
               rowRequestSource = ('web' if len(row[5]) > 5 else 'api')
               request_count['web'] += (1 if rowRequestSource == 'web' else 0)
               request_count['api'] += (1 if rowRequestSource == 'api' else 0)
            
            if rowGET['idp']:
                if ('idp' in criteria and rowGET['idp'].find(criteria['idp']) >= 0) or 'idp' not in criteria:
                    if rowGET['idp'] not in idp.keys():
                        idp[rowGET['idp']] = 0
                    idp[rowGET['idp']] += 1
                    request_param['idp'] += 1
                   
            if rowGET['reg_auth']:
                if ('reg_auth' in criteria and rowGET['reg_auth'].find(criteria['reg_auth']) >= 0) or 'reg_auth' not in criteria:
                    if rowGET['reg_auth'] not in reg_auth.keys():
                        reg_auth[rowGET['reg_auth']] = 0
                    reg_auth[rowGET['reg_auth']] += 1
                    request_param['reg_auth'] += 1

            result[rowDate] = {
                'request_count' : request_count,
                'request_param' : request_param,
                'request_uniq' : {'idp':len(idp),'reg_auth':len(reg_auth)},
                'idp' : idp,
                'reg_auth' : reg_auth
            }
        
    return result

# Parse URL from log line. Used to get only idp and reg_auth. 
def parseReqURL(url):
   result = {'idp':False,'reg_auth':False}
   parsed_url = urlparse(url)
   captured_value = parse_qs(parsed_url.query)

   if 'idp' in captured_value:
      result['idp'] = captured_value['idp'][0]
   
   if 'reg_auth' in captured_value:
      result['reg_auth'] = captured_value['reg_auth'][0]
   
   return result

def filterParsedData(json_data,criteria):
   idp = {}
   reg_auth = {}
   out_data = {}
   
   if criteria['cur_date'] in json_data: 
      if 'idp' in criteria and len(criteria['idp']): 
         for idpEntry in json_data[criteria['cur_date']]['idp'].keys():
            if idpEntry.find(criteria['idp']) >= 0:
               idp[criteria['idp']] = json_data[criteria['cur_date']]['idp'][criteria['idp']]
      else:
         idp = json_data[criteria['cur_date']]['idp']

      if 'reg_auth' in criteria and len(criteria['reg_auth']):
         for regAuthEntry in json_data[criteria['cur_date']]['reg_auth'].keys():
            if regAuthEntry.find(criteria['reg_auth']) >= 0:
               reg_auth[criteria['reg_auth']] = json_data[criteria['cur_date']]['reg_auth'][criteria['reg_auth']]
      else:
         reg_auth = json_data[criteria['cur_date']]['reg_auth']
   
      out_data[criteria['cur_date']] = {
            'request_count' : json_data[criteria['cur_date']]['request_count'],
            'request_param' : json_data[criteria['cur_date']]['request_param'],
            'request_uniq' : {'idp':len(idp),'reg_auth':len(reg_auth)},
            'idp' : idp,
            'reg_auth' : reg_auth,
      }
      
   return out_data

### Classes

# /api/test
class Test(Resource):
    def get(self):
        return {'test':'It Works!'}


# /api/eccsresults
class EccsResults(Resource):
    def get(self):

       file_path = f"{e_p.ECCS_OUTPUTDIR}/{e_p.ECCS_RESULTSLOG}"
       date = e_p.DAY
       status = None
       idp = None
       reg_auth = None
       eccsDataTable = False
       simple = False
       check_result = None

       if 'eccsdt' in request.args:
          eccsDataTable = True
       if 'date' in request.args:
          date = request.args['date']
          file_path = f"{e_p.ECCS_OUTPUTDIR}/eccs_{date}.log"
       if 'status' in request.args:
          status = request.args['status'].upper()
          if (status not in ['OK','DISABLED','ERROR','UNKNOWN']):
              return jsonify(error="Incorrect status provided. It can be 'OK','DISABLED','ERROR','UNKNOWN'")
       if 'idp' in request.args:
          idp = request.args['idp']
          if (not existsInFile(file_path, idp, "entityID", eccsDataTable, date)):
             return jsonify(error=f"Identity Provider not found with the entityID: {idp}")
       if 'reg_auth' in request.args:
          reg_auth = request.args['reg_auth']
          if (not existsInFile(file_path, reg_auth, "registrationAuthority", eccsDataTable, date)):
             return jsonify(error=f"Identity Providers not found with the Registration Authority: {reg_auth}")
       if 'format' in request.args:
          if 'simple' == request.args['format']:
             simple = True
       if 'check_result' in request.args:
          check_result = request.args['check_result']
          if (check_result not in ['OK','Timeout','Unable-To-Check','Connection-Error','No-SP-Metadata-Error','SSL-Error','IdP-Generic-Error','403-Forbidden','DISABLED']):
              return jsonify(error="Incorrect check_result value provided. It can be 'OK','Timeout','Unable-To-Check','Connection-Error','No-SP-Metadata-Error','SSL-Error','IdP-Generic-Error', '403-Forbidden' or 'DISABLED'")

       lines = []
       results = []
       try:
          with open(file_path,"r",encoding="utf-8") as fo:
               lines = fo.readlines()

       except FileNotFoundError as e:
           if (eccsDataTable):
              return ''
           else:
              return jsonify(error=f'FileNotFound: ECCS script has not been executed on {date}')

       for line in lines:
          # Strip the line feed and carriage return characters
          line = line.rstrip("\n\r")

          # Loads the json line into aux
          aux = json.loads(line)
    
          aux['date'] = date

          # If the results are for ECCS DataTable, otherwise... remove only "mailto:" prefix
          if (eccsDataTable):
             aux['contacts']['technical'] = buildEmailAddress(aux['contacts']['technical'])
             aux['contacts']['support'] = buildEmailAddress(aux['contacts']['support'])
          else:
             aux['contacts']['technical'] = aux['contacts']['technical'].replace("mailto:","")
             aux['contacts']['support'] = aux['contacts']['support'].replace("mailto:","")

          if (idp and status):
              if (idp == aux['entityID'] and status == aux['status']):
                 if (simple):              
                    auxSimple = getSimpleDict(aux)
                    results.append(auxSimple)
                 else:
                    results.append(aux)
          elif (reg_auth and status):
              if (reg_auth == aux['registrationAuthority'] and status == aux['status']):
                 if (simple):              
                    auxsimple = getSimpleDict(aux)
                    results.append(auxsimple)
                 else:
                    results.append(aux)
          elif (reg_auth and check_result):
              if (reg_auth == aux['registrationAuthority'] and (check_result == aux['sp1']['checkResult'] or check_result == aux['sp2']['checkResult'])):
                 if (simple):              
                    auxsimple = getSimpleDict(aux)
                    results.append(auxsimple)
                 else:
                    results.append(aux)
          elif (status and check_result):
              if (status == aux['status'] and (check_result == aux['sp1']['checkResult'] or check_result == aux['sp2']['checkResult'])):
                 if (simple):              
                    auxsimple = getSimpleDict(aux)
                    results.append(auxsimple)
                 else:
                    results.append(aux)
          elif (idp):
              if (idp == aux['entityID']):
                 if (simple):              
                    auxSimple = getSimpleDict(aux)
                    results.append(auxSimple)
                 else:
                    results.append(aux)
          elif (reg_auth):
              if (reg_auth == aux['registrationAuthority']):
                 if (simple):              
                    auxSimple = getSimpleDict(aux)
                    results.append(auxSimple)
                 else:
                    results.append(aux)
          elif (status):
              if (status == aux['status']):
                 if (simple):              
                    auxSimple = getSimpleDict(aux)
                    results.append(auxSimple)
                 else:
                    results.append(aux)
          elif (check_result):
              if (check_result == aux['sp1']['checkResult'] or check_result == aux['sp2']['checkResult']):
                 if (simple):              
                    auxSimple = getSimpleDict(aux)
                    results.append(auxSimple)
                 else:
                    results.append(aux)
          elif (simple):
              auxSimple = getSimpleDict(aux)
              results.append(auxSimple)

          else:
             results.append(aux)

       return jsonify(results)


# /api/fedstats
class FedStats(Resource):
   def get(self):
       list_feds = get_list_from_url(e_p.ECCS_LISTFEDSURL, e_p.ECCS_LISTFEDSFILE)
       regAuthDict = get_reg_auth_dict(list_feds)

       file_path = f"{e_p.ECCS_OUTPUTDIR}/{e_p.ECCS_RESULTSLOG}"
       date = e_p.DAY
       reg_auth = None
       eccsDataTable = False

       if ('date' in request.args):
          date = request.args['date']
          file_path = f"{e_p.ECCS_OUTPUTDIR}/eccs_{date}.log"
       if ('reg_auth' in request.args):
          reg_auth = request.args['reg_auth']
          if (not existsInFile(file_path, reg_auth, "registrationAuthority", eccsDataTable, date)):
             return jsonify(error="Registration Authority not found")

       lines = []
       results = []
       try:
          with open(file_path,"r",encoding="utf-8") as fo:
               lines = fo.readlines()

       except FileNotFoundError as e:
           if (eccsDataTable):
              return ''
           else:
              return jsonify(error=f'FileNotFound: ECCS script has not been executed on {date} yet')

       if (reg_auth):
          resultDict = {'date': date, 'registrationAuthority': reg_auth, 'OK': 0, 'ERROR': 0, 'DISABLED': 0, 'UNKNOWN': 0}

          for line in lines:
              # Strip the line feed and carriage return characters
              line = line.rstrip("\n\r")

              # Loads the json line into aux
              aux = json.loads(line)

              if (aux['registrationAuthority'] == reg_auth):
                 if (aux['status'] == "OK"):
                    resultDict['OK'] = resultDict['OK'] + 1
                 if (aux['status'] == "ERROR"):
                    resultDict['ERROR'] = resultDict['ERROR'] + 1
                 if (aux['status'] == "DISABLED"):
                    resultDict['DISABLED'] = resultDict['DISABLED'] + 1
                 if (aux['status'] == "UNKNOWN"):
                    resultDict['UNKNOWN'] = resultDict['UNKNOWN'] + 1

          results.append(resultDict)
          return jsonify(results)

       else:
           for name,regAuth in regAuthDict.items():
               resultDict = {'date': date, 'registrationAuthority': regAuth, 'OK': 0, 'ERROR': 0, 'DISABLED': 0, 'UNKNOWN': 0}

               for line in lines:
                   # Strip the line feed and carriage return characters
                   line = line.rstrip("\n\r")

                   # Loads the json line into aux
                   aux = json.loads(line)

                   if (regAuth == aux['registrationAuthority']):
                      if (aux['status'] == "OK"):
                         resultDict['OK'] = resultDict['OK'] + 1
                      if (aux['status'] == "ERROR"):
                         resultDict['ERROR'] = resultDict['ERROR'] + 1
                      if (aux['status'] == "DISABLED"):
                         resultDict['DISABLED'] = resultDict['DISABLED'] + 1
                      if (aux['status'] == "UNKNOWN"):
                         resultDict['UNKNOWN'] = resultDict['UNKNOWN'] + 1

               results.append(resultDict)
           return jsonify(results)

        
# /api/webdata
class WebData(Resource):
   def get(self):
      file_path = f"{e_p.ECCS_LOGSDIR}/eccs-uwsgi-req.log" # will this name be moved to properties definer file ?
      criteria = {}
      criteria['date_from'] = criteria['date_to'] = e_p.DAY
      # TBM to config
      eccsLogRotated = True
      useParsedFile = False
      
      in_data = request.args

      if ('dateFrom' in in_data and isValidDate(in_data['dateFrom'])):
         criteria['date_from'] = in_data['dateFrom']
         if ('dateTo' not in in_data):
            criteria['date_to'] = (datetime.strptime(criteria['date_from'], '%Y-%m-%d') + timedelta(days=30)).strftime('%Y-%m-%d')
            
            if datetime.today().strftime('%Y-%m-%d') < criteria['date_to']:
               diff = (datetime.strptime(criteria['date_to'], '%Y-%m-%d') - datetime.today()).days
               criteria['date_from'] = (datetime.strptime(criteria['date_from'], '%Y-%m-%d') - timedelta(days=diff)).strftime('%Y-%m-%d')
               criteria['date_to'] = datetime.today().strftime('%Y-%m-%d')
      
      if ('dateTo' in in_data and isValidDate(in_data['dateTo'])):
         criteria['date_to'] = in_data['dateTo']
         if ('dateFrom' not in in_data):
            criteria['date_from'] = (datetime.strptime(criteria['date_to'], '%Y-%m-%d') + timedelta(days=30)).strftime('%Y-%m-%d')
      
      if ('requestSource' in in_data and in_data['requestSource'] == 'divided'):
         criteria['request_source'] = 'divided'

      if 'regAuth' in in_data:
         criteria['reg_auth'] = unquote(in_data['regAuth'])
      
      if 'idp' in in_data:
         criteria['idp'] = unquote(in_data['idp'].strip())

      lines = []
      results = {}
      cur_date = criteria['date_from']

      if eccsLogRotated == True:
         while cur_date <= criteria['date_to']:
            json_data = {}
            criteria['logfile_date'] = (datetime.strptime(cur_date, '%Y-%m-%d') + timedelta(days=1)).strftime('%Y-%m-%d')
            criteria['cur_date'] = cur_date
            tmpDate = datetime.strptime(criteria['logfile_date'], '%Y-%m-%d').strftime('%Y%m%d')
            file_path = f"{e_p.ECCS_LOGSDIR}/eccs-uwsgi-req.log-{tmpDate}"

            if useParsedFile == True:
               json_file_path = f"{e_p.ECCS_DIR}/parsed/eccs-uwsgi-req-json-{tmpDate}"
               try:
                  f = open(json_file_path)
                  json_data = json.load(f)

               except (ValueError, FileNotFoundError) as e:
                  #print(e)
                  pass
               
               if len(json_data) > 0 and ('idp' in criteria or 'reg_auth' in criteria):
                  json_data = filterParsedData(json_data, criteria)

            if len(json_data) == 0:
               try:
                  with open(file_path,"r",encoding="utf-8") as fo:
                     lines = fo.readlines()

                  json_data = parseLog(lines, criteria)

                  if useParsedFile == True and 'idp' not in criteria and 'reg_auth' not in criteria:
                     storeParsedDay(json_file_path, json_data)

               except FileNotFoundError as e:
                  # print(e)
                  pass
            
            results.update(json_data)
            cur_date = (datetime.strptime(cur_date, '%Y-%m-%d') + timedelta(days=1)).strftime('%Y-%m-%d')
      else:
            try:
               with open(file_path,"r",encoding="utf-8") as fo:
                  lines = fo.readlines()

            except FileNotFoundError as e:
               # print(e)
               results = {}

            results = parseLog(lines, criteria)

      return json.dumps(results)

# /api/
class Help(Resource):
    def get(self):
        return { 'ECCS JSON Interface': 'https://wiki.geant.org/display/eduGAIN/eduGAIN+Connectivity+Check+2#eduGAINConnectivityCheck2-JSONinterface' }

@app.route('/getsamlreq')
def getSamlReq():
    # Setup list_eccs_idps
    list_eccs_idps = get_list_from_url(e_p.ECCS_LISTIDPSURL, e_p.ECCS_LISTIDPSFILE)

    entityid_idp = request.args.get('idp')
    entityid_sp = request.args.get('sp')

    for idp in list_eccs_idps:
        if (idp['entityID'] == entityid_idp):
           for sp in e_p.ECCS_SPS:
               if (sp['entityID'] == entityid_sp):
                  return f"<meta http-equiv='Refresh' content='0; url={generate_login_url(sp['entityID'], sp['http_post_acs_location'], idp['Location'])}' />"
               else:
                  fake_sp_acs_url = f"{entityid_sp.split('shibboleth')[0]}Shibboleth.sso/SAML2/POST"
                  return f"<meta http-equiv='Refresh' content='0; url={generate_login_url(entityid_sp,fake_sp_acs_url, idp['Location'])}' />"


# Routes
api.add_resource(Help, '/') # Route_1
api.add_resource(Test, '/test') # Route_2
api.add_resource(EccsResults, '/eccsresults') # Route_3
api.add_resource(FedStats, '/fedstats') # Route_4
api.add_resource(WebData, '/webdata') # Route_4

if __name__ == '__main__':

   # Useful only for API development Server
   #app.config['JSON_AS_ASCII'] = True
   #app.logger.removeHandler(default_handler)
   #app.logger = get_logger("eccs_api.log", e_p.ECCS_LOGSDIR, "w", "INFO")
   app.run(port='5002')
